/*
Copyright (c) 2014, Intel Corporation

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
 * sends an endless stream of Layer 2 packets out of
 * <interface> that cycles through all 10 PTP V2 packet
 * types.
 *
 * Modified for testing purposes by
 *		David Ertman <davidx.m.ertman@intel.com>
 *
 * Based off of the program "timestamping" by
 *		Patrick Ohly <patrick.ohly@intel.com>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. * See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>

#include <sys/time.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <net/ethernet.h>
#include <netpacket/packet.h>

#include <asm/types.h>

#include "packets.h"

/* #define PTP_LENGTH 31 */

#define ERR_ARGUMENTS 1
#define ERR_SETUP 2
#define ERR_TX_TIMESTAMP 4
#define ERR_RX_TIMESTAMP 8

int pVersion;

static void usage(const int error)
{
	printf("send_L2 Version 1.2.3\n\n");
	printf("\n\nUSAGE:\n");
	printf("send_L2 [-V|--pVersion [2]] <interface>\n\n");
	printf("where interface is a valid ethernet device\n");
	printf("-V|-pVersion [2]\tUsed to limit transmit to only PTP V2 SYNC packets\n\n");
	exit(error);
}

static void bail(const char *error)
{
	printf("%s: %s\n", error, strerror(errno));
	exit(ERR_SETUP);
}

/* trap the alarm signal so that pause() will wake up on receipt */
static void handle_alarm(int s)
{
	return;
}

static int install_handler(int signum, void(*handler)(int))
{
	struct sigaction action;
	sigset_t mask;

	/* Unblock the signal */
	sigemptyset(&mask);
	sigaddset(&mask, signum);
	sigprocmask(SIG_UNBLOCK, &mask, NULL);

	/* Install the signal handler */
	action.sa_handler = handler;
	action.sa_flags = 0;
	sigemptyset(&action.sa_mask);
	sigaction(signum, &action, NULL);

	return 0;
}

static void sendpacket(int sock, struct sockaddr *addr, socklen_t addr_len)
{
	int res;
	int x, y; /* loop variable */

	for (x = 0; x < 14; x++) {
		y = (pVersion == 0) ? x : 0;

		res = 0;
		if ((x < 4) || (x > 7)) {
			while (res == 0)
				res = sendto(sock, packets_v2[y], sizeof(packets_v2[y]), 0, addr, addr_len);
			if (res < 0)
				fprintf(stderr, "%s: %s\n", "send", strerror(errno));
			usleep(200);
		}
	}
}

int main(int argc, char **argv)
{
	char *interface;
	int sock;
	const unsigned char ether_broadcast_addr[] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
	struct ifreq device;
	struct sockaddr_ll addr;
	struct timeval next;

	pVersion = 0;

	while (1) {
		static struct option long_options[] = {
			{"pVersion", required_argument, 0, 'V'},
			{"help", no_argument,       0, 'h'},
			{0, 0, 0, 0}
		};

		int option_index = 0;
		int c;

		c = getopt_long (argc, argv, "V:h",
				 long_options, &option_index);

		/* no more options found */
		if (c == -1)
			break;

		switch (c) {
		case 'V':
			pVersion = 2;
			break;
		case 'h':
			usage(0);
		case '?':
		case '0':
		default:
			usage(ERR_ARGUMENTS);
		}

	}

	if (optind >= argc)
		usage(ERR_ARGUMENTS);
	interface = argv[optind];

	sock = socket(PF_PACKET, SOCK_DGRAM, htons(0x88f7));
	if (sock < 0)
		bail("socket");

	memset(&device, 0, sizeof(device));
	strncpy(device.ifr_name, interface, sizeof(device.ifr_name));
	if (ioctl(sock, SIOCGIFINDEX, &device) < 0)
		bail("getting interface index");

	addr.sll_family = AF_PACKET;
	addr.sll_ifindex = device.ifr_ifindex;
	addr.sll_halen = ETHER_ADDR_LEN;
	addr.sll_protocol = htons(0x88f7);
	memcpy(addr.sll_addr, ether_broadcast_addr, ETHER_ADDR_LEN);

	if (bind(sock,
		 (struct sockaddr *)&addr,
		 sizeof(addr)) < 0)
		bail("bind");

	install_handler(SIGALRM, handle_alarm);

	/* send packets forever every second */
	gettimeofday(&next, 0);
	next.tv_sec = (next.tv_sec + 1);
	next.tv_usec = 0;
	while (1) {
		struct timeval now;
		struct itimerval timer;
		long delta_us;

		gettimeofday(&now, 0);
		delta_us = (long)(next.tv_sec - now.tv_sec) * 1000000 +
			(long)(next.tv_usec - now.tv_usec);
		if (delta_us > 0) {
			memset(&timer, 0, sizeof(timer));
			timer.it_value.tv_usec = delta_us;
			setitimer(ITIMER_REAL, &timer, NULL);
			pause();
		} else {
			/* write one packet */
			sendpacket(sock,
				   (struct sockaddr *)&addr,
				   sizeof(addr));
			printf("Sent data at %ld.%09lu or %s",
			       now.tv_sec, now.tv_usec,
			       ctime(&now.tv_sec));
			next.tv_sec++;
			continue;
		}
	}

	return 0;
}
